// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "PrefabTool.hpp"

using namespace Dawn;


struct PrefabComponentParserRuleset PrefabRegistry::getRuleset(std::string type) {
  std::vector<std::string> pathsToScan;
  pathsToScan.push_back(this->sources);

  auto it = pathsToScan.begin();
  while(it != pathsToScan.end()) {
    auto str = *it;
    pathsToScan.erase(it);

    Directory dir = Directory(str);
    auto children = dir.readDirectory();
    
    auto itChildren = children.begin();
    while(itChildren != children.end()) {
      if(itChildren->second == DIRECTORY_CHILD_TYPE_DIRECTORY) {
        pathsToScan.push_back(str + "/" + itChildren->first);
      } else {
        if(itChildren->first == type+".hpp") {
          // Load ruleset
          std::string data;
          File file(str + "/" + itChildren->first);
          if(!file.readString(&data)) {
            std::cout << "Failed to read ruleset file!" << file.filename << std::endl;
          } else {
            // Begin scanning contentsr
            // std::string include = this->sources;
            auto toRemove = File::normalizeSlashes(this->sources + "/");
            auto includePath = file.filename.substr(toRemove.size(), file.filename.size() - toRemove.size());

            // Now locate the first subdir since we don't want to include the root path (e.g. dawn, dawnrose, etc)
            auto firstSlash = includePath.find("/");
            if(firstSlash != std::string::npos) {
              includePath = includePath.substr(firstSlash + 1, includePath.size() - firstSlash - 1);
            }
            
            struct PrefabComponentParserRuleset ruleset;
            ruleset.include = includePath;
            ruleset.name = type;

            // Find each instance of "@optional" when it's used within a comment
            // e.g. // @optional or /* @optional */ in the string data.
            std::regex regex("^(\\s)*(\\/\\/|\\/\\*\\s*)(\\@optional)(\\s*\\*\\/)?(\\s)$", std::regex_constants::ECMAScript);
            std::sregex_iterator it(data.begin(), data.end(), regex);
            std::sregex_iterator end;

            while(it != end) {
              std::smatch match = *it;

              auto nextLineStart = data.rfind('\n', match.position() + match.length()) + 1;
              auto nextLineEnd = data.find('\n', nextLineStart);
              auto nextLine = data.substr(nextLineStart, nextLineEnd - nextLineStart);

              std::regex regex2("^\\s*([^\\s]+)\\s+([^\\s;]+)\\s*;", std::regex_constants::ECMAScript);
              std::sregex_iterator it2(nextLine.begin(), nextLine.end(), regex2);
              std::sregex_iterator end2;
              while(it2 != end2) {
                // Log out all matches
                std::smatch match2 = *it2;
                std::string type = match2[1].str();
                std::string name = match2[2].str();

                // Type may be StateProperty<type>, so we need to extract the type
                auto regex3 = std::regex("^StateProperty<([^>]+)>$", std::regex_constants::ECMAScript);
                std::smatch match3;
                if(std::regex_match(type, match3, regex3)) {
                  type = match3[1].str();
                }

                // Set a parser for this type
                ruleset.optional[name] = type;
                ++it2;
              }
              ++it;
            }
            return ruleset;
          }
        }
      }
      ++itChildren;
    }
    it = pathsToScan.begin();
  };

  return { .name = "" };
}

//

std::vector<std::string> PrefabTool::getRequiredFlags() {
  return { "input", "output", "sources" };
}

std::map<std::string, std::string> PrefabTool::getOptionalFlags() {
  return std::map<std::string, std::string>();
}

int32_t PrefabTool::start() {
  File input = File(flags["input"]);
  if(!input.exists()) {
    std::cout << "Input file does not exist!" << std::endl;
    return 1;
  }

  std::string data;
  if(!input.readString(&data)) {
    std::cout << "Failed to read input file!" << std::endl;
    return 1;
  }

  struct PrefabRegistry registry;
  registry.sources = flags["sources"];

  auto xml = Xml::load(data);
  std::string error;
  struct Prefab prefab;
  prefab.registry = &registry;
  auto result = ((PrefabParser()).parse(&xml, &prefab, &error));
  if(result != 0) {
    std::cout << "Failed to parse prefab: " << error << std::endl;
    return result;
  }

  // Generate output
  std::vector<std::string> outputData;
  PrefabGen::generate(&outputData, &prefab, "");

  // Load output file from name and type
  File outputFile = File(flags["output"] + "/prefabs/" + prefab.type + "/" + prefab.name + ".hpp");
  if(!outputFile.mkdirp()) {
    std::cout << "Failed to create output directory!" << std::endl;
    return 1;
  }

  // Combine vector into single string
  std::string outputDataStr = "";
  auto it = outputData.begin();
  while(it != outputData.end()) {
    outputDataStr += *it + "\n";
    ++it; 
  }
  
  if(!outputFile.writeString(outputDataStr)) {
    std::cout << "Failed to write output file!" << std::endl;
    return 1;
  }

  return 0;
}

//

std::vector<std::string> PrefabParser::getRequiredAttributes() {
  return { "name", "type" };
}

std::map<std::string, std::string> PrefabParser::getOptionalAttributes() {
  return { };
}

int32_t PrefabParser::onParse(
  Xml *node,
  std::map<std::string, std::string> values,
  struct Prefab *out,
  std::string *error
) {
  out->name = values["name"];
  out->type = values["type"];

  auto itChildren = node->children.begin();
  while(itChildren != node->children.end()) {
    // Parse children as components
    struct PrefabComponent component;
    component.registry = out->registry;

    auto ret = (PrefabComponentParser()).parse(*itChildren, &component, error);
    if(ret != 0) return ret;
    out->components.push_back(component);
    ++itChildren;
  }

  return 0;
}

//

std::vector<std::string> PrefabComponentParser::getRequiredAttributes() {
  return { };
}

std::map<std::string, std::string> PrefabComponentParser::getOptionalAttributes() {
  return { { "ref", "" } };
}

int32_t PrefabComponentParser::onParse(
  Xml *node,
  std::map<std::string, std::string> values,
  struct PrefabComponent *out,
  std::string *error
) {
  // Check if values contains key "ref" and has a string of size more than 0
  if(values.find("ref") != values.end() && values["ref"].size() > 0) {
    out->ref = values["ref"];
  }

  // Get the ruleset.
  auto ruleset = out->registry->getRuleset(node->node);
  if(ruleset.name.size() == 0) {
    *error =  "Unknown prefab node type: " + node->node;
    return 1;
  }

  out->include = ruleset.include;
  out->type = node->node;

  // Define parser types here.
  auto rawParser = [&](std::string v){
    return v;
  };

  auto stringParser = [&](std::string v){
    return "\"" + v + "\"";
  };

  auto floatParser = [&](std::string v){
    // Make sure number contains decimal
    if(v.find(".") == std::string::npos) {
      v += ".0";
    }
    // Make sure number contains a number before the decimal
    if(v.find(".") == 0) {
      v = "0" + v;
    }
    // Make sure ends with f
    if(v.find("f") == std::string::npos) {
      v += "f";
    }
    return v;
  };


  // Iterate all the optional attributes and store within the out if present
  auto itOptional = ruleset.optional.begin();
  while(itOptional != ruleset.optional.end()) {
    // Determine the parser to use
    auto type = itOptional->second;

    std::function<std::string(std::string)> parser = rawParser;
    if(type.find("string") != std::string::npos) {
      parser = stringParser;
    } else if(type.find("float") != std::string::npos) {
      parser = floatParser;
    } else if(type.find("*") == (type.size() - 1)) {
      type = type.substr(0, type.size() - 1);
      parser = rawParser;
    } else {
      *error = "Unkown parser for type: " + type;
      return 1;
    }

    if(node->attributes.find(itOptional->first) != node->attributes.end()) {
      auto raw = node->attributes[itOptional->first];
      out->values[itOptional->first] = parser(raw);
    }
    ++itOptional;
  }

  return 0;
}

//

void PrefabGen::generate(
  std::vector<std::string> *out,
  struct Prefab *info,
  std::string tabs
) {
  struct ClassGenInfo classInfo;
  classInfo.clazz = info->name;
  classInfo.extend = "SceneItemPrefab<" + info->name + ">";
  classInfo.constructorArgs = "Scene *scene, sceneitemid_t id";
  classInfo.extendArgs = "scene, id";

  struct MethodGenInfo methodAssets;
  methodAssets.name = "prefabAssets";
  methodAssets.isStatic = true;
  methodAssets.type = "std::vector<Asset*>";
  methodAssets.args = "AssetManager *man";
  line(&methodAssets.body, "std::vector<Asset*> assets;", "");

  struct MethodGenInfo methodInit;
  methodInit.name = "prefabInit";
  methodInit.isOverride = true;
  methodInit.args = "AssetManager *man";

  classInfo.includes.push_back("prefab/SceneItemPrefab.hpp");

  // Includes
  int32_t componentNumber = 0;
  auto itChildren = info->components.begin();
  while(itChildren != info->components.end()) {
    auto c = *itChildren;
    std::string name = "c" + std::to_string(componentNumber++);
    bool_t init = true;
    if(c.ref.size() > 0) {
      init = false;
      name = c.ref;
      line(&classInfo.publicProperties, c.type + " *" + c.ref + ";", "");
    }

    // Initialize
    line(&methodInit.body, (init ? "auto " : "") + name + " = this->addComponent<" + c.type + ">();", "");

    // Now set each property
    auto itValues = c.values.begin();
    while(itValues != c.values.end()) {
      line(&methodInit.body, name + "->" + itValues->first + " = " + itValues->second + ";", "");
      ++itValues;
    }
    line(&methodInit.body, "", "");

    classInfo.includes.push_back(c.include);
    ++itChildren;
  }
  
  // Seal methods
  line(&methodAssets.body, "return assets;", "");

  // Add in methods
  CodeGen::methodGen(&classInfo.publicCode, methodAssets);
  line(&classInfo.publicCode, "", "");
  CodeGen::methodGen(&classInfo.publicCode, methodInit);

  CodeGen::classGen(out, classInfo);
}