// Copyright (c) 2023 Dominic Masters
// 
// This software is released under the MIT License.
// https://opensource.org/licenses/MIT

#include "DawnHostVita.hpp"
#include "game/DawnGame.hpp"

using namespace Dawn;

DawnHost::DawnHost() {

}

int32_t DawnHost::init(DawnGame *game) {
  this->game = game;

  vglInit(0x800000);
  vglWaitVblankStart(GL_TRUE);

  // Override the defaults
  game->renderManager.backBuffer.setSize(DAWN_VITA_WIDTH, DAWN_VITA_HEIGHT);

  // Initialize the game
  auto result = game->init();
  if(result != DAWN_GAME_INIT_RESULT_SUCCESS) return result;

  // Hard-Load the first scene assets.
  if(game->scene != nullptr) {
    auto assets = game->scene->getRequiredAssets();
    game->assetManager.queueLoad(assets);
    game->assetManager.syncLoad();
    game->scene->stage();
  }

  return DAWN_HOST_INIT_RESULT_SUCCESS;
}

int32_t DawnHost::start(DawnGame *game) {
  double_t time, newTime;
  float_t fDelta;
  int32_t updateResult;

  // Main Render Loop 
  time = 0.0f;
	for(;;) {
    // Determine the delta.
    float_t fDelta = 1.0f / 60.0f;
    time = newTime;

    // Perform update
    updateResult = this->update(game, fDelta);

    // Did the update complete successfully?
    if(updateResult == DAWN_HOST_UPDATE_RESULT_EXIT) {
      break;
    } else if(updateResult != DAWN_HOST_UPDATE_RESULT_SUCCESS) {
      return updateResult;
    }
    
		// Performing buffer swap
		vglSwapBuffers(GL_FALSE);
  }

  return DAWN_HOST_START_RESULT_EXIT_SUCCESS;
}

int32_t DawnHost::update(DawnGame *game, float_t delta) {
  // Tick game.
  auto ret = game->update(delta);
  switch(ret) {
    case DAWN_GAME_UPDATE_RESULT_SUCCESS:
      return DAWN_HOST_UPDATE_RESULT_SUCCESS;
    
    case DAWN_GAME_UPDATE_RESULT_EXIT:
      return DAWN_HOST_UPDATE_RESULT_EXIT;

    default:
      return ret;
  }
}

void DawnHost::unload(DawnGame *game) {
  vglEnd();
}

DawnHost::~DawnHost() {
}


int main(int argc, char **args) {
  int32_t result;

  // Create the host
  auto host = new DawnHost();
  auto game = new DawnGame(host);

  // Initialize the host and error check
  result = host->init(game);
  switch(result) {
    case DAWN_HOST_INIT_RESULT_SUCCESS:
      break;
    default:
      return result;
  }

  // Request the main loop to start running.
  result = host->start(game);
  switch(result) {
    case DAWN_HOST_START_RESULT_SUCCESS:
      break;
    case DAWN_HOST_START_RESULT_EXIT_SUCCESS:
      break;
    default:
      return result;
  }

  // Main loop finished without errors, cleanup
  host->unload(game);

  delete game;
  delete host;

  // Success
  return 0;
}